-- ------ --
-- Player --
-- ------ --

local GravityEntity           = require 'entities.GravityEntity'
local Block                   = require 'entities.Block'
local Monster                 = require 'entities.Monster'
local Hitbox                  = require 'entities.Hitbox'
local Lockon                  = require 'entities.Lockon'
local Object                  = require 'entities.Object'
local damage                  = require 'data.damage'
local cron                    = require 'lib.cron'
local bump                    = require 'lib.bump'
local map                     = require 'map'

-- local wasd                    = true
local wasd                    = false
local Commands                =    {}   --  store all commands here

if wasd then
  Commands["moveLeft"]        =   "a"
  Commands["moveRight"]       =   "d"
  Commands["moveUp"]          =   "w"
  Commands["moveDown"]        =   "s"
  Commands["actJump"]         =   "j"
  Commands["actAttack"]       =   "k"
  Commands["actSpecial"]      =   "l"
  Commands["actLockon"]       =   " "
else
  Commands["moveLeft"]        =   "s"
  Commands["moveRight"]       =   "f"
  Commands["moveUp"]          =   "e"
  Commands["moveDown"]        =   "d"
  Commands["actJump"]         =   "j"
  Commands["actAttack"]       =   "k"
  Commands["actSpecial"]      =   "i"
  Commands["actLockon"]       =   "l"
end

local jCommands               =    {}   --  joystick commands
  jCommands["moveLeft"]       =   {1,-1}
  jCommands["moveRight"]      =   {1,1}
  jCommands["moveUp"]         =   {2,-1}
  jCommands["moveDown"]       =   {2,1}
  -- jCommands["actJump"]        =   2
  jCommands["actJump"]        =   11
  -- jCommands["actAttack"]      =   1
  jCommands["actAttack"]      =   13
  -- jCommands["actSpecial"]     =   4
  jCommands["actSpecial"]     =   14
  -- jCommands["actLockon"]      =   3
  jCommands["actLockon"]      =   12

local loffset,toffset         =   0,0   -- for hitbox placement

local Face                    =  true   -- which direction we're facing (default right)

local Player                  = class('Player', GravityEntity)

local E = {1,2,3,4,5}
for k,v in ipairs(E) do
  E[k] = getElement(k)
end

local gd = globalDamage                 --  damage formula table

local runAccel                =   800   --  player acceleration while going left/right
local breakAccel              =  1200   --  player acceleration when stopping/turning around
local jumpVelocity            =   800   --  initial upwards velocity when jumping
local minVelocity             =    10   --  velocity at which the player stops moving

local function sign(x)
  return x < 0 and -1 or (x > 0 and 1 or 0)
end

local function abs(x)
  return x < 0 and -x or x
end

local function acc(x, y, z) -- x init, y max, z degree
  if abs(x-y) < z then return y end
  return x + sign(z)
end

function Player:initialize(x,y,n,p)
  GravityEntity.initialize(self,x,y,32,64)
  self.name                   =     n
  self.underFeet              =    {}
  self.canFly                 = false
  self.score                  =    {}   --  a table of numeric values for the player
    self.score["hits"]        =     0   --  total number of hits
    self.score["kills"]       =     0   --  total number of kills
    self.score["exp"]         =     0   --  total number of experience
    self.score["deaths"]      =     0   --  total number of deaths
    self.score["life"]        =   100   --  current health
    self.score["power"]       =   100   --  current power
    self.score["cost"]        =    10   --  power cost for special moves
    self.score["pps"]         =     2   --  power recovered per second
    self.score["combo"]       =     0   --  power recovered per second
    self.score["jtimer"]      =     0   --  duration of current jump
    self.score["maxjump"]     =  0.25   --  maximum jump duration
    self.score["maxrun"]      =   320   --  maximum run speed
    self.score["maxlockon"]   =  1024   --  maximum lockon distance
    self.score["maxlife"]     =   100   --  full health
    self.score["maxpower"]    =   100   --  full power
    self.score["level"]       =     1   --  current level
    self.score["lvllife"]     =    10   --  amount of life gained each level
    self.score["lvlpower"]    =     5   --  amount of power gained each level
  self.status       =              {}   --  a table of boolean values for the player
    self.status["hitbox"]     = false   --  checks if we have an active hitbox
    self.status["neutral"]    = false   --  unused
    self.status["dying"]      = false   --  checks if we've been hit
    self.status["armor"]      = false   --  checks if we are currently invulnerable
    self.status["bounce"]     = false   --  checks if we need to rebound on hit
    self.status["crouch"]     = false   --  checks if we are crouching
    self.status["killer"]     = false   --  checks if we just made a kill
    self.status["hitter"]     = false   --  checks if we just made a hit
    self.status["powerup"]    = false   --  checks if we are powered up
    self.status["flipping"]   = false   --  checks if we are doing a flip
    self.status["special"]    = false   --  checks if we are doing a special move
    self.status["lockon"]     = false   --  checks if we have an active lockon
    self.status["dash"]       = false   --  checks if we are doing a dash
    self.status["earth"]      = false   --  checks if we can make elemental earth attacks
    self.status["plant"]      = false   --  checks if we can make elemental plant attacks
    self.status["water"]      = false   --  checks if we can make elemental water attacks
    self.status["fire"]       = false   --  checks if we can make elemental fire attacks
    self.status["metal"]      = false   --  checks if we can make elemental metal attacks
    self.status["coma"]       = false   --  checks if we are in a sushi coma
    self.status["movelock"]   = false   --  checks if movement is currently locked
    self.status["friction"]   = false   --  checks for locked moves with friction
    self.status["airlock"]    = false   --  checks if we should recover when we hit the ground
    self.status["facelock"]   = false   --  checks if we are stuck facing a direction
    self.status["xlock"]      = false   --  checks for x-axis lock
    self.status["ylock"]      = false   --  checks for y-axis lock
  self.buffer       =   {
    one     =   {C=false,T=0},
    two     =   {C=false,T=0},
    three   =   {C=false,T=0},
    four    =   {C=false,T=0},
    five    =   {C=false,T=0} } 
  self.element                = false   --  store current element
  self.message                = false   --  store status message
  self.myhitbox               = false   --  variable containing the currently active hitbox
  self.lvlup                  = false   --  we set this when gaining levels
  if p then                             --  called initialize with a score argument
    self.score = p
  end
  initKill(self,self.score["hits"],self.score["kills"])
end

function Player:change()                --  does nothing, use later for negative status
end

function Player:getStatus(s)            --  get a specific status bit
  return self.status[s]
end

function Player:allStatus()             --  get all bits in the status table
  return self.status
end

function Player:makehitbox(x,y,t,p,c)     --  t=type,p=powerup
  local hw,hh = 32,32
  KTS = getKillType(t,"s")              --  shape
  KTD = getKillType(t,"d")              --  base damage
  KTP = getKillType(t,"p")              --  affected by powerup
  if KTS == "wide" then
    hw,hh = 64,32
  elseif KTS == "high" then
    hw,hh = 32,64
  elseif KTS == "square" then
    hw,hh = 48,48
  elseif KTS == "splash" then
    hw,hh = 96,48
  end
  -- if p then                             --  when powered up, increase hitbox size
    if (hw > hh) then
      hw = hw * 2
    elseif (hh > hw) then
      hh = hh * 2
    else
      -- hw,hh = hw*2,hh*2
    end
  if p then                             --  when powered up, increase damage
    if KTP then KTD = KTD * 2 end
  end
  if c then
    Hitbox:initialize(x,y,hw,hh,self,self.element,KTD,true)
  else
    Hitbox:initialize(x,y,hw,hh,self,self.element,KTD)
  end
  self.status["hitbox"] = true
  self.status["bounce"] = getKillType(t,"b")  -- front-load bounce state, will execute on hit
end

function Player:movehitbox(l,t)         --  move hitbox 
  Hitbox:move(l,t)
end

function Player:gethitbox()             --  returns hitbox associated with player
  return Hitbox:retrieve(self)
end

function Player:destroyhitbox(this)     --  remove hitbox
  local hi = self:gethitbox()
  if hi then
    hi:remove(self)
    hi:destroy()
  end
  self.status["hitbox"] = false
  self.status["bounce"] = false
end

function Player:makelockon(x,y,d)       --  make lockon
  if d then                             --  d for direction
    -- d = 8                               --  comment this out for 4-way lockon
    Lockon:initialize(x,y,16,16,self,d) 
    if (d == 8) or (d == 2) then
      self.status["lockon"] = "X"
    elseif (d == 4) or (d == 6) then
      self.status["lockon"] = "Y"
    else
      self.status["lockon"] = true
    end
  else
    Lockon:initialize(x,y,16,16,self)
    self.status["lockon"] = true
  end
  self.mylockon = self:getlockon()
end

function Player:getlockon()
  return Lockon:retrieve(self)
end

function Player:getlockonhit()
  return Lockon:gethit(self)
end

function Player:movelockon(l,t)
  Lockon:move(l,t)
end

function Player:destroylockon(this)
  local lo = self:getlockon()
  lo:remove(self)
  lo:destroy()
  self.status["lockon"] = false
  self.distance = false
end

function Player:shouldCollide(other)
  -- if self.status.flipping and (other == self.flipbox) then
  if self.status.armor then             --  we pass through enemies when armored
    return instanceOf(Block, other) or instanceOf(Object, other)
  elseif self.element and other.element and (self.element == other.element) then
    return instanceOf(Block, other) or instanceOf(Object, other)
  else
    return instanceOf(Block, other) or instanceOf(Monster, other) or instanceOf(Object, other)
  end
end

function Player:collision(other, dx, dy)
  if not other.damage then other.damage=10 end
  if instanceOf(Monster, other) then
    if not self.status["dying"] then
      self.status["dying"] = true
      if not self.status["armor"] then
        if self.element and other.element and (checkElement(other.element,self.element)) then
          self:takedamage(other.damage,2)
        else
          self:takedamage(other.damage)
        end
        self.score["deaths"] = self.score["deaths"] + 1
      end
    end
  elseif instanceOf(Object, other) then
    if (other.n.c == "sushi") then
      self.score["life"]=math.min(self.score["life"]+other.n.life,self.score["maxlife"])
      self.score["power"]=math.min(self.score["power"]+other.n.power,self.score["maxpower"])
      if other.n.t then
        self.status[other.n.t] = true
      end
      if other.n.s then
        if other.n.s == "coma" then
          self.status["movelock"] = true
          self.status["facelock"] = true
          self.status["friction"] = true
          self.status["coma"] = true
          self.message = "sushi coma"
          cron.tagged('coma').after(2, function() self.status["movelock"] = false self.status["coma"] = false self.status["facelock"] = false self.status["friction"] = false self.message=false end)
        end
      end
      self:doPowerUp()
      -- self.status["bounce"] = "y"
    elseif (other.n.c == "weapon") then
    elseif (other.n.c == "special") then
      if other.n.a == "lvlup" then self.lvlup = true end
      if other.n.a == "mapup" then map.lvlup = true end
    end
    other:destroy()
  elseif dx~=0 or dy~=0 then            --  we hit a block, reset velocity
    if dx~=0 and sign(self.vx) ~= sign(dx) then self.vx = 0 end
    if dy~=0 and sign(self.vy) ~= sign(dy) then self.vy = 0 end
     
    if other.type=="lava" then          --  the block is a hazard
      if not self.status["dying"] and not self.status["armor"] then
        self.status["dying"] = true
        self.status["armor"] = true
        self:takedamage(gd["lava"].b*gd["lava"].l*map.level)
        self.score["deaths"] = self.score["deaths"] + 1
      end
    end

    if dy < 0 then                      --  we are on the ground
      self.underFeet[other] = true
      self.flipbox = other
      if self:getlockon() then self:destroylockon(self) end
      if not love.keyboard.isDown(Commands["actJump"]) then 
        self.score["jtimer"] = 0       --  only reset this if jump has been released
      end
      if self.status["airlock"] then
        self.status["movelock"] = false
        self.status["airlock"] = false
        if self:gethitbox() then self:destroyhitbox(self) end
        self.status["friction"] = false
      end
      -- if other.type=="lava" then
      --   if not self.status["dying"] and not self.status["armor"] then
      --     self.status["dying"] = true
      --     self.status["armor"] = true
      --     self:takedamage(other.damage)
      --     self.score["deaths"] = self.score["deaths"] + 1
      --   end
      -- end
    end
    -- update the player position so that the intersection stops occurring
    self.l, self.t = self.l + dx, self.t + dy
  end
end

function Player:endCollision(block)
  self.underFeet[block] = nil
end

function Player:isOnGround()
  for _,_ in pairs(self.underFeet) do
    return true
  end
  return false
end

function Player:populateBuffer(c)       --  not implemented but populate anyway
  self.buffer["five"] = self.buffer["four"]
  self.buffer["four"] = self.buffer["three"]
  self.buffer["three"] = self.buffer["two"]
  self.buffer["two"] = self.buffer["one"]
  self.buffer["one"] = {C=c,T=os.time()}
end

function Player:checkBuffer(c)          --  not implemented
  local b = self.buffer
  local T = os.time()
  if (T - b.three.T < 2) then
    if c == "upper" then 
      if (b.one.C == "moveRight" and b.two.C == "moveDown" and b.three.C == "moveRight") then
        return true
      elseif (b.one.C == "moveLeft" and b.two.C == "moveDown" and b.three.C == "moveLeft") then
        return true
      end
    elseif c == "tackle" then
      -- if (b.one.C == "moveRight" and b.two.C == "moveDown" and b.three.C == "moveLeft") then
      if (b.one.C == "moveRight" and b.two.C == "moveDown") then
        return true
      -- elseif (b.one.C == "moveLeft" and b.two.C == "moveDown" and b.three.C == "moveRight") then
      elseif (b.one.C == "moveLeft" and b.two.C == "moveDown") then
        return true
      end
    end
  else
    return false
  end
end

-- not-implemented update snippet, check buffer and execute if command matches
-- if not s.armor then
--   if cc.actAttack and not co.actAttack then
--     if ((self:checkBuffer("upper") or self:checkBuffer("tackle")) and n.power >= 5) then
--       if (cc.moveRight or cc.moveLeft) then
--         if cc.moveDown then
--           vx,vy = self:doUpper()
--         else
--           vx,vy = self:doTackle()
--         end
--         n.power = math.max(n.power - 5, 0)
--       end
--     end
--   end
-- end

function Player:takedamage(d,m)         --  d=damage value, m=multiplier
  cron.tagged('hitbox').cancel()
  self.status["armor"] = true
  self.status["crouch"] = false
  self.damagetimer = cron.tagged('hurt').after(1, function() self:doRecovery() end)
  if m then
    self.score["life"] = math.max((self.score["life"] -(d * m)), 0)
  else
    self.score["life"] = math.max((self.score["life"] -d), 0)
  end
  self.score["combo"] = 0
end

function Player:dying()                 --  call once after being hit
  if self:getlockon() then self:destroylockon(self) end
  if self:gethitbox() then self:destroyhitbox(self) end
  self.status["movelock"] = true
  self.status["airlock"] = true
  self.status["dying"] = false
  self.status["friction"] = false
  self.status["xlock"] = false
  self.status["ylock"] = false
  if Face then -- facing right
    vx,vy = -self.score["maxrun"] ,(-jumpVelocity/2)
  else -- facing left
    vx,vy = self.score["maxrun"] ,(-jumpVelocity/2)
  end
  return vx,vy
end

function Player:doRecovery()
  self.status["movelock"] = false
  self.status["friction"] = false
  self.status["facelock"] = false
  self.status["armor"] = false
  self.status["xlock"] = false
  self.status["ylock"] = false
end

function Player:resetStatus()
  if self:gethitbox() then self:destroyhitbox(self) end
  if self:getlockon() then self:destroylockon(self) end
  self.distance = false
  for k,v in pairs(self.status) do
    self.status[k] = false
  end
  -- self.vx, self.vy = 0,0
end

function Player:gainLevel()
  self.score["level"] = self.score["level"] + 1
  self.message = "level " .. self.score["level"]
  cron.tagged('level').cancel()
  cron.tagged('level').after(5, function() self.message=false end)
  self.score["maxlife"] = self.score["maxlife"] + self.score["lvllife"]
  self.score["life"] = math.min(self.score["life"] + self.score["lvllife"],self.score["maxlife"])
  self.score["maxpower"] = self.score["maxpower"] + self.score["lvlpower"]
  self.score["power"] = math.min(self.score["power"] + self.score["lvlpower"],self.score["maxpower"])
end

-- ------- --
-- Actions --
-- ------- --

function Player:doHeal(d)               --  heal self (special)
  self.status["armor"] = true
  self.status["movelock"] = true
  self.status["facelock"] = true
  self.status["friction"] = true
  cron.tagged('hitbox').cancel()
  cron.tagged('hurt').cancel()
  self.healtimer = cron.tagged('hurt').after(2, function() self:doRecovery() if self:gethitbox() then self:destroyhitbox(self) end if self:getlockon() then self:destroylockon(self) end end)
  self.score["life"] = math.min(self.score["life"] + 10,self.score["maxlife"])
end

function Player:doPowerUp(d)            --  power up weapon attacks (special)
  self.status["powerup"] = true
  cron.tagged('powerup').cancel()
  cron.tagged('powerup').after(10, function() self.status["powerup"] = false end)
end

function Player:doTackle(d)             --  tackle (special)
  local vx, vy = self.vx, self.vy
  local l,t = self.l,self.t
  local f = Face
  toffset = -8
  loffset = -8
  self.status["ylock"] = false
  if d == 1 then
    vx,vy = 1.5*-self.score["maxrun"], jumpVelocity
  elseif d == 3 then
    vx,vy = 1.5*self.score["maxrun"], jumpVelocity
  elseif d == 7 then
    vx,vy = 1.5*-self.score["maxrun"], -jumpVelocity
  elseif d == 9 then
    vx,vy = 1.5*self.score["maxrun"], -jumpVelocity
  elseif d == 4 then
    self.status["ylock"] = true
    vx,vy = 2*-self.score["maxrun"],0
  elseif d == 6 then
    self.status["ylock"] = true
    vx,vy = 2*self.score["maxrun"],0
  end
 
  cron.tagged('hitbox').cancel()
  cron.tagged('hurt').cancel()
  if self:getlockon() then self:destroylockon(self) end
  if self:gethitbox() then self:destroyhitbox(self) end
  self.status["crouch"] = true
  self.status["armor"] = true
  self.status["movelock"] = true
  self.status["dash"] = true
  self:makehitbox((l+loffset),(t+toffset),"tackle")
  self.tackletimer = cron.tagged('hitbox').after(0.5, function() self:doRecovery() self.vx = 0 if self:gethitbox() then self:destroyhitbox(self) end self.status["dash"] = false end)
  return vx,vy
end

function Player:doUpper(d)              --  uppercut (special)
  local vx, vy = self.vx, self.vy
  local l,t = self.l,self.t
  local f = Face
  if d then
    toffset = -32
    loffset = 0
    vx, vy = 0 ,-jumpVelocity
  else
    -- toffset = -16
    toffset = -32
    if f then -- facing right
      vx, vy = self.score["maxrun"] ,-jumpVelocity
      loffset = 16
    else -- facing left
      vx, vy = -self.score["maxrun"] ,-jumpVelocity
      loffset = -16
    end
  end
  cron.tagged('hitbox').cancel()
  cron.tagged('hurt').cancel()
  if self:getlockon() then self:destroylockon(self) end
  if self:gethitbox() then self:destroyhitbox(self) end
  self:makehitbox((l+loffset),(t+toffset),"upper")
  self.status["movelock"] = true
  self.status["friction"] = true
  self.status["airlock"] = true
  self.status["armor"] = true
  self.status["crouch"] = false
  self.uppertimer = cron.tagged('hitbox').after(0.8, function() self:doRecovery() if self:gethitbox() then self:destroyhitbox(self) end end)
  return vx,vy
end

function Player:doSplash(d)             --  splash shockwave (special)
  local vx,vy = self.vx, self.vy
  local f = Face
  local l,t = self.l,self.t
  -- toffset = -8
  toffset = -8
  loffset = -80
  self.status["ylock"] = false
  cron.tagged('hitbox').cancel()
  cron.tagged('hurt').cancel()
  if self:getlockon() then self:destroylockon(self) end
  if self:gethitbox() then self:destroyhitbox(self) end
  self.status["crouch"] = true
  self.status["armor"] = true
  self.status["movelock"] = true
  self:makehitbox((l+loffset),(t+toffset),"splash",false,true)
  self.tackletimer = cron.tagged('hitbox').after(0.5, function() self:doRecovery() self.vx = 0 if self:gethitbox() then self:destroyhitbox(self) end end)
  return vx,vy
end

function Player:doDive(d)               --  executes a divekick
  local vx, vy = self.vx, self.vy
  local l,t = self.l,self.t
  local f = Face
  if d then                             --  vertical divekick
    toffset = 48
    loffset = -8
    vx, vy = 0, jumpVelocity
  else                                  --  angled divekick
    toffset = 48
    if f then -- facing right
      vx, vy = self.score["maxrun"] ,jumpVelocity
      loffset = 0
    else -- facing left
      vx, vy = -self.score["maxrun"] ,jumpVelocity
      loffset = -16
    end
  end
  if self:getlockon() then self:destroylockon(self) end
  if self:gethitbox() then self:destroyhitbox(self) end
  self:doRecovery()
  cron.tagged('hitbox').cancel()
  self:makehitbox((l+loffset),(t+toffset),"dive")
  self.status["movelock"] = true
  self.status["friction"] = false
  self.status["airlock"] = true
  -- self.divetimer = cron.tagged('player').after(0.5, function() self:doRecovery() end)
  return vx,vy
end

function Player:doSlide(d)              --  executes a sliding kick (from the ground)
  local l,t = self.l,self.t
  local f = Face
  toffset = 0
  if f then
    vx = self.score["maxrun"]
    loffset = 32
  else
    vx = -self.score["maxrun"]
    loffset = -32
  end
  if self:gethitbox() then self:destroyhitbox(self) end
  self:makehitbox((l+loffset),(t+toffset),"slide")
  self.status["facelock"] = true
  self.status["movelock"] = true
  self.status["crouch"] = true
  self.slideactive = cron.tagged('hitbox').after(0.25, function() self.status["friction"] = true end)
  self.sliderecovery = cron.tagged('hitbox').after(0.5, function() self:doRecovery() if self:gethitbox() then self:destroyhitbox(self) end end)
  return vx,vy
end

function Player:doSlash(d)              --  executes a standard attack
  local l,t = self.l,self.t
  local f = Face
  local p = self.status["powerup"]
  local htype = "slash"
  if d then                             --  vertical slash
    htype = "vslash"
    toffset = -48
    -- if d=="down" then toffset = 48 end
    if d=="down" then toffset = 16 end
    if f then loffset = 16 else loffset = -16 end
    -- if p then 
      if d=="down" then
      else
        toffset = toffset - 64
      end
    -- end
  else                                  --  horizontal slash
    toffset = 0
    if f then
      loffset = 16
    else
      loffset = -48
      -- if p then
        loffset = loffset - 64
      -- end
    end
  end
  if self:gethitbox() then self:destroyhitbox(self) end
  self:makehitbox((l+loffset),(t+toffset),htype,p)
  self.status["facelock"] = true
  if self:isOnGround() then self.status["friction"] = true end
  self.attacktimer = cron.tagged('hitbox').after(0.25, function() self:doRecovery() if self:gethitbox() then self:destroyhitbox(self) end end)
end

function Player:getSwing()
  local ssx,ssy = self:getCenter()
  local lo = self:getlockon()
  local lsx,lsy = lo:getCenter()
  return math.sqrt( (ssx-lsx)^2 + (ssy-lsy)^2 ),ssx,ssy
end

function Player:doSwing(self, dt, maxdt)
  local ml,mt,mvx,mvy,g = GravityEntity.update(self, dt, maxdt, true)
  local n = self.score
  local ssx,ssy = self:getCenter()
  local lo = self:getlockon()
  local lsx,lsy = lo:getCenter()
  local msx,msy = ml + math.abs(ssx - self.l), mt + math.abs(ssy - self.t)
  local dx,dy = msx - lsx, msy - lsy
  local d = math.sqrt(dx^2 + dy^2)
  local f = d/self.distance
  local x,y = lsx + dx / f, lsy + dy / f
  local l,t = x - math.abs(self.l - ssx), y - math.abs(self.t - ssy)
  local vx,vy = (x - ssx) / dt, (y - ssy) / dt
  vy = vy + (g * dt)
  return l,t,vx,vy
end

function Player:update(dt, maxdt)
  -- temporary elemental alignment toggling
  if love.keyboard.isDown("1") then
    if self:getStatus(E[1]) then self.element = E[1] end
  elseif love.keyboard.isDown("2") then
    if self:getStatus(E[2]) then self.element = E[2] end
  elseif love.keyboard.isDown("3") then
    if self:getStatus(E[3]) then self.element = E[3] end
  elseif love.keyboard.isDown("4") then
    if self:getStatus(E[4]) then self.element = E[4] end
  elseif love.keyboard.isDown("5") then
    if self:getStatus(E[5]) then self.element = E[5] end
  elseif love.keyboard.isDown("0") then
    self.element = false
  end

  -- destroy the player if we run out of life
  if (self.score["life"] < 1) then
    if self.getlockon() then self.destroylockon() end
    if self.gethitbox() then self.destroyhitbox() end
    self:destroy()
  end

  -- increment kill counter if we get a kill
  if self.score["kills"] < getKill(self) then
    self.status["killer"] = true
    self.score["kills"] = getKill(self)
    self.score["exp"] = self.score["exp"] + (1 * map.level)
    local ssl = self.score["level"]
    local exptolevel = ((ssl^2)+4)^2
    if (math.fmod(self.score["exp"],exptolevel) == 0) then
      self.lvlup = true
    end
  else
    self.status["killer"] = false
  end

  -- increment hit counter if we get a hit
  if self.score["hits"] < getHit(self) then
    self.status["hitter"] = true
    self.score["hits"] = getHit(self)
  else
    self.status["hitter"] = false
  end

  -- combo stuff
  if self:isOnGround() then
    self.score["combo"] = 0
  end

  self.myhitbox = self:gethitbox()      --  current hitbox
  self.mylockon = self:getlockon()      --  current lockon

  -- if self:getlockon() then              --  make sure this is set if we have a lockon
  --   self.distance=self:getSwing()
  -- end

  -- local variables
  local vx,vy = self.vx,self.vy
  local l,t = self.l,self.t
  local h,w = self.h,self.w
  local f,c = Face,Commands
  local s = self.status
  local n = self.score
  local b = self.buffer
  local jc = jCommands

  -- commands
  if cc then co = cc else co = {} end   --  store previous commands, or empty
  cc = {}                               --  clear current command table
  for k,v in pairs(c) do                --  populate with current commands
    if love.keyboard.isDown(v) then
      cc[k] = true
      if not co[k] then
        if string.find(k,'move') then
          self:populateBuffer(k)        --  update the command buffer
        end
      end
    else
      cc[k] = false
    end
  end
  local getjoy = love.joystick.getJoysticks()
  local joy = getjoy[1]
  if joy then
  for k,v in pairs(jc) do
    if string.find(k,'move') then
      if v[2] > 0 then
        if joy:getAxis(v[1]) >= (v[2]/2) then cc[k] = true end
      elseif v[2] < 0 then
        if joy:getAxis(v[1]) <= (v[2]/2) then cc[k] = true end
      -- if joy:getAxis(v[1]) == v[2] then cc[k] = true end
      end
    else
      if joy:isDown(v) then cc[k] = true end
    end
  end
  end

  if s.dying then                       --  being hit
    vx,vy = self:dying()
  end

  if not s.movelock then                --  allowed to move

    s.special = false                   --  reset when movelock ends

    if cc.moveDown then                 --  crouch
      if self:isOnGround() then
        s.crouch = true
      end
    else
      if not s.facelock then
        s.crouch = false
      end
    end

    if cc.moveLeft then                 --  move left
      if not s.facelock then Face = false end
      if s.crouch or (s.facelock and self:isOnGround()) then
        vx = math.max((vx - dt * 5 * (vx > 0 and breakAccel or runAccel)), (-n.maxrun/4))
      else
        vx = math.max((vx - dt * 5 * (vx > 0 and breakAccel or runAccel)), -n.maxrun)
      end
    elseif cc.moveRight then            --  move right
      if not s.facelock then Face = true end
      if s.crouch or (s.facelock and self:isOnGround()) then
        vx = math.min((vx + dt * 5 * (vx < 0 and breakAccel or runAccel)), (n.maxrun/4))
      else
        vx = math.min((vx + dt * 5 * (vx < 0 and breakAccel or runAccel)), n.maxrun)
      end
    else                                --  brake until stoping
      vx = vx - dt * breakAccel * sign(vx)
      if abs(vx) < minVelocity then vx = 0 end
    end

    if cc.actJump then                  --  jump was pressed
      if not co.actJump then            --  it wasn't held down
        if self:isOnGround() then
          if cc.moveDown and not s.facelock then
            vx = self:doSlide()
          else
            vy = -jumpVelocity
          end
        else                            --  not on the ground so max the timer
          n.jtimer = n.maxjump          --  auto max if we let go of the button
          if cc.moveLeft or cc.moveRight then  -- angled divekick
            if cc.moveDown and not self:getlockon() then vx,vy = self:doDive() end
          else                          --  vertical divekick
            if cc.moveDown and not self:getlockon() then vx,vy = self:doDive(true) end
          end  
        end
      else                              --  jump is still pressed
        if vy >= 0 then                 --  hit a ceiling, max timer
          n.jtimer = n.maxjump
        else                            --  increment timer until max
          n.jtimer = n.jtimer + dt
        end
        if n.jtimer < n.maxjump then    --  if we still have time, keep rising
          vy = -jumpVelocity
        end
      end
    -- elseif co.actJump then              --  we just let go the jump button
    --   if n.jtimer < n.maxjump then
    --     n.jtimer = n.maxjump
    --     vy = vy/2
    --   end
    end

    if cc.actLockon and not co.actLockon then                --  lockon is pressed
      if not s.lockon then              --  no current lockon, toggle on
        if (n.power >= 1) and not self:isOnGround() then
          if cc.moveUp and not cc.moveLeft and not cc.moveRight then
            self:makelockon(self.l+8,self.t,8)
          -- elseif cc.moveDown then self:makelockon(self.l+8,self.t+64,2)
          elseif cc.moveRight and not cc.moveLeft then
            if cc.moveUp then
              self:makelockon(self.l+32,self.t,8)
            else
              -- self:makelockon(self.l+32,self.t,6)
              self:makelockon(self.l+32,self.t,9)
            end
          elseif cc.moveLeft and not cc.moveRight then
            if cc.moveUp then
              self:makelockon(self.l,self.t,8)
            else
              -- self:makelockon(self.l,self.t,4)
              self:makelockon(self.l,self.t,7)
            end
          else
            self:makelockon(self.l+8,self.t,8)
          end
          if not s.powerup then n.power = math.max(n.power - 1, 0) end
        end                             --  activate only in the air and subtract cost
      else                              --  lockon exists, toggle off
        if self:getlockon() then self:destroylockon(self) end
      end
    else                                --  lockon not pressed, remove lockons
    --  if self:getlockon() then self:destroylockon(self) end
    end

    if cc.actAttack then                --  do an attack
      if not co.actAttack and not s.facelock then
        if cc.moveUp then
          self:doSlash(true)            --  anti-air (vertical) attack
        elseif cc.moveDown and not self:isOnGround() then
          self:doSlash("down")
        else
          self:doSlash()                --  standard (horizontal) attack
        end
      end
    end
            
  else                                  --  movement other than gravity while movelocked

    n.jtimer = n.maxjump                --  max the jump timer right away

    if s.friction then                  --  inertia
      vx = vx - dt * breakAccel * sign(vx)
      if abs(vx) < minVelocity then vx = 0 end
    end

    if not cc.actSpecial then           --  reset only if key is released while in movelock
      s.special = false
    end

    if s.dash and self:isOnGround() then
      vx,vy = self:doSplash()
      s.friction = true
      s.dash = false
    end

    if s.hitter then                    --  we just made a hit
      if s.bounce then
        if s.bounce == "y" then         --  y-axis bounce (vertical)
          vy = -jumpVelocity
        elseif s.bounce == "x" then     --  x-axis bounce (horizontal)
          if f then
            vx = -n.maxrun
          else
            vx = n.maxrun
          end
        elseif s.bounce == "v" then     --  variable bounce (into tackle)
          if cc.actSpecial and not s.special and (n.power >= n.cost) then
            n.power = math.max(n.power - n.cost, 0)
            s.special = true
            if f then
              Face = false
              if cc.moveUp and not cc.moveDown then
                vx,vy = self:doTackle(7)
              elseif cc.moveDown and not cc.moveUp then
                vx,vy = self:doTackle(1)
              else
                vx,vy = self:doTackle(4)
              end
            else
              Face = true
              if cc.moveUp and not cc.moveDown then
                vx,vy = self:doTackle(9)
              elseif cc.moveDown and not cc.moveUp then
                vx,vy = self:doTackle(3)
              else
                vx,vy = self:doTackle(6)
              end
            end
          end
        end
        if s.bounce == "x" or s.bounce == "y" then
          if self:getlockon() then self:destroylockon(self) end
          if self:gethitbox() then self:destroyhitbox(self) end
          self:doRecovery()
        end
      end
    end

  end

  -- we can cancel anything into a special regadless of movelock
  if not s.armor or (s.armor and not s.movelock) then
    if cc.actSpecial and not co.actSpecial then
      if (n.power >= n.cost) then
        s.special = true
        if cc.moveRight and not cc.moveUp and not cc.moveDown then
          vx,vy = self:doTackle(6)
        elseif cc.moveLeft and not cc.moveUp and not cc.moveDown then
          vx,vy = self:doTackle(4)
        elseif cc.moveRight and cc.moveUp and not cc.moveDown then
          -- vx,vy = self:doTackle(9)
          vx,vy = self:doTackle(6)
        elseif cc.moveRight and cc.moveDown and not cc.moveUp then
          vx,vy = self:doTackle(3)
        elseif cc.moveLeft and cc.moveDown and not cc.moveUp then
          vx,vy = self:doTackle(1)
        elseif cc.moveLeft and cc.moveUp and not cc.moveDown then
          -- vx,vy = self:doTackle(7)
          vx,vy = self:doTackle(4)
        elseif cc.moveUp and not cc.moveLeft and not cc.moveRight then
          vx,vy = self:doUpper()
        elseif cc.moveDown and not cc.moveLeft and not cc.moveRight then
          if f then
            vx,vy = self:doTackle(3)
          else
            vx,vy = self:doTackle(1)
          end
        else
          if f then
            vx,vy = self:doTackle(6)
          else
            vx,vy = self:doTackle(4)
          end
        end
        if not s.powerup then n.power = math.max(n.power - n.cost, 0) end
      end
    end
  else                                  --  except if we are already armored
  end
  -- SPLASH
  if s.xlock then vx = 0 end            --  lock x-axis movement
  if s.ylock then vy = -28 end          --  lock y-axis movement (approximate vs. gravity)

  if s.crouch then                      --  resize the bbox if we are crouching
    if self:isOnGround() or s.movelock then
      if h == 64 then
        t = t + 32
      end
      h = 32
    else
      s.crouch = false
    end
  else
    if h == 32 then
      t = t -32
    end
    h = 64
  end

  if self:getlockon() and self:getlockonhit() then  --  stop falling if the lockon hits
    if not self.distance then self.distance = self:getSwing() end
    if cc.actJump and not co.actJump then   --  check if we're re-jumping from lockon
      if self:getlockon() then self:destroylockon(self) end
      if self:gethitbox() then self:destroyhitbox(self) end
      vy = -jumpVelocity
      self.vy = vy
      GravityEntity.update(self, dt, maxdt)
    else                                --  not jumping
      local lon = self:getlockon()
      local ltar = lon.target
      local ssx,ssy = self:getCenter()
      local ssl = self.score["maxlockon"]
      if ltar.type and ltar.type=="slick" then  --  check for invalid targets
        self:destroylockon(self)
      else                              --  lockon target is good, do swing stuff
        s.jtimer = s.maxjump            --  first we max the jump timer
        if cc.moveRight then            --  add right velocity
          if ssx > lon.l and ssy > (lon.t + lon.h) then
            self.l = (self.l + (jumpVelocity/4)*dt)
          end
        elseif cc.moveLeft then         --  add left velocity
          if ssx < (lon.l + lon.w) and ssy > (lon.h + lon.t) then
            self.l = (self.l - (jumpVelocity/4)*dt)
          end
        elseif cc.moveUp then           --  reel in the line
          if self.distance > self.h then
            self.distance = self.distance - (jumpVelocity/4)*dt
          end
        elseif cc.moveDown then         --  let out the line
          if self.distance < ssl then
            self.distance = math.min(self.distance + (jumpVelocity/4)*dt,ssl-dt)
          end
        end                             --  main swing routine
        self.l,self.t,self.vx,self.vy = self:doSwing(self, dt, maxdt)
      end
    end
  else                                  -- normal gravity, main routine
    self.vx, self.vy = vx, vy
    self.l,self.t,self.w,self.h = l,t,w,h
    GravityEntity.update(self, dt, maxdt)
  end

  if self.status["hitbox"] == true then self:movehitbox((self.l+loffset),(self.t+toffset)) end 
  if s.lockon then  
    -- re-evaluate this, uncertain if we should drag the lockon or fire and forget
    -- if (s.lockon == "X") then
      -- self.mylockon:move(self.l+8,self.mylockon.t)
    -- elseif (s.lockon == "Y") then
      -- self.mylockon:move(self.mylockon.l,self.t)
    -- else
    --   if f then
        -- self.mylockon:move(self.l+8+self.mylockon.l,self.mylockon.t)
    --   else
        -- self.mylockon:move(self.mylockon.l-self.l+8,self.mylockon.t)
     --  end
    -- end
    Lockon:update(self.mylockon,dt,maxdt)
    --  limit max lockon distance
    if self.distance and self.distance > n.maxlockon then self:destroylockon(self) end
  else -- only recover power when not armored and not using lockon
    if not s.armor then n.power = math.min(n.power + n.pps * dt,n.maxpower) end
  end

end

function Player:draw()
  local r,g,b = 0,255,0
  if self.status["armor"] then
    r,g,b = 255,0,0
  elseif self.status["powerup"] then
    r,g,b = 0,0,255
  end
  love.graphics.setColor(r,g,b,100)
  love.graphics.rectangle('fill', self:getBBox())
  love.graphics.setColor(r,g,b)
  love.graphics.rectangle('line', self:getBBox())
  local bl,bt,bw,bh = self:getBBox()
  if self.element and self:gethitbox() then
    love.graphics.setColor(255,255,255)
    love.graphics.setFont(bigFont)
    love.graphics.print(string.upper(string.sub(self.element,1,1)),bl,bt-32)
    love.graphics.setFont(defaultFont)
  end
  if Face then bl = bl + 16 end
  bw,bh = 16,16
  -- love.graphics.setColor(255,100,0)
  love.graphics.setColor(255,255,255)
  if self:getlockon() then
    local drawl = self:getlockon()
    local lpx,lpy = drawl:getCenter()
    love.graphics.line(lpx,lpy, self:getCenter())
  end
  -- love.graphics.rectangle('line',bl,bt,bw,bh)
  love.graphics.print("@",bl,bt)
  if self.message then 
    love.graphics.print(self.message,self.l,self.t-16)
  end
end

return Player
