require 'lib.middleclass'
math.randomseed( os.time() )
local bump        = require 'lib.bump'
local bump_debug  = require 'lib.bump_debug'
local camera      = require 'lib.camera'
local cron        = require 'lib.cron'

local menu        = require 'menu'
local map         = require 'map'
local Entity      = require 'entities.Entity'
local Player      = require 'entities.Player'
local Object      = require 'entities.Object'

local calendar    = require 'data.calendar'

local maxdt       = 0.1      --  if the window loses focus/etc, use this instead of dt
local fts         = 120      --  for fixed timestep stuff
local updater     = {}       --  updater table for fixed timestep
local drawDebug   = false    --  draw bump debug info
local gamePaused  = false    --  pause the game
local gameStart   = false    --  start menu
local gameType    = false    --  store the game type
local player                 --  a reference to the player (so the camera can follow)
local diagnostics = [[

    mem:   %dKb
    dt:     %f
    fps:   %d
    x:     %d
    y:     %d
    vx:     %d
    vy:     %d
    
]]

love.window.setTitle("BENTO")

-- ---- --
-- Bump --
-- ---- --

function bump.collision(obj1, obj2, dx, dy)
  obj1:collision(obj2,  dx,  dy)
  obj2:collision(obj1, -dx, -dy)
end

function bump.endCollision(obj1, obj2)
  obj1:endCollision(obj2)
  obj2:endCollision(obj1)
end

function bump.shouldCollide(obj1, obj2)
  return obj1:shouldCollide(obj2) or
         obj2:shouldCollide(obj1)
end

function bump.getBBox(obj)
  return obj:getBBox()
end

-- ---- --
-- Load --
-- ---- --

local function reset(t,p)
  resetDamage()
  if map.level then
    map.level=map.level+1
  else
    map.level=1
  end
  if hardreset then
    map.level=map.level-1
    hardreset=false
  end
  if p then
    map.reset(t,map.level)
    player:resetStatus()
    player = Player:new(60, map.height-100,"bento",p)
  else
    map.reset(t,map.level)
    player = Player:new(60, map.height-100,"bento")
  end
  camera.setBoundary(0,0,map.width,map.height)
  camera.setViewport(0,0,love.graphics.getWidth(),love.graphics.getHeight())
end

function love.load()
  defaultFont = love.graphics.newFont(12)
  bigFont = love.graphics.newFont(32)
  updater.max = (1 / fts)
  updater.now = 0
  -- screensize = love.window.setMode(1200,720, {fullscreen=normal, vsync=true, fsaa=0})
  -- screensize = love.window.setMode(969,540, {fullscreen="desktop", vsync=true, fsaa=0})
  screensize = love.window.setMode(0,0, {fullscreen="desktop", vsync=true, fsaa=0})
  gameStart = true
end

-- ------ --
-- Update --
-- ------ --

local timebank = 0
function love.update(dt)
  mapHeight = map.height
  dt = math.min(dt, maxdt)

  timebank = timebank + dt
  if timebank >= 12 then                --  advance the calendar every 12 seconds
    advanceCalendar()
    timebank = 0
  end

  updater.now = updater.now + dt        --  put things we want at fixed dt here
  if updater.now >= updater.max then
    updateDamage()
    updater.now = 0
  end

  cron.update(dt)
  -- camera.lookAt(player:getFeet())    --  camera doesn't jump when we crouch
  if gameStart then                     --  start menu
    if love.keyboard.isDown("x") then
      reset("x")
      gameStart = false
      gameType = "x"
    elseif love.keyboard.isDown("c") then
      reset("c")
      gameStart = false
      gameType = "c"
    elseif love.keyboard.isDown("v") then
      reset("v")
      gameStart = false
      gameType = "v"
    elseif love.keyboard.isDown("m") then
      reset("m")
      gameStart = false
      gameType = "m"
    elseif love.keyboard.isDown("b") then
      reset("b")
      gameStart = false
      gameType = "b"
    end
  elseif gamePaused then                --  halt all movement if game is paused
  else
    camera.lookAt(player:getFeet())       --  camera doesn't jump when we crouch
    -- local l,t,w,h = camera.getViewport() -- update only the screen
    local l,t,w,h = 0,0,map.width,map.height -- update the whole world
    local updateEntity = function(entity) entity:update(dt, maxdt) end
    bump.each(updateEntity, l,t,w,h)
    bump.collide(l,t,w,h)
    if player.lvlup then
      player:gainLevel()
      player.lvlup = false
    end
    if map.lvlup then
      player:resetStatus()
      reset(gameType,player.score)
      map.lvlup = false
    end
  end

end

-- ------- --
-- Drawing --
-- ------- --

local function drawEntity(entity) entity:draw() end
local function drawCameraStuff(l,t,w,h)
  if drawDebug then bump_debug.draw(l,t,w,h) end
  bump.each(drawEntity, l,t,w,h)
end

function drawLifebar(p,m)
  love.graphics.setColor(50,255,50,200)
  love.graphics.rectangle('fill', 10,10,(p * 2),15)
  love.graphics.setColor(255,255,255)
  love.graphics.rectangle('line', 10,10,(m * 2),15)
end

function drawPowerbar(p,m)
  love.graphics.setColor(50,50,255,200)
  love.graphics.rectangle('fill', 10,30,math.floor(p * 2),15)
  love.graphics.setColor(255,255,255)
  love.graphics.rectangle('line', 10,30,math.floor(m * 2),15)
end

function drawGameOver()
  love.graphics.setColor(255,0,0)
  love.graphics.setFont(bigFont)
  -- love.graphics.print("GAME OVER",384,300)
  love.graphics.print("GAME OVER",480,320)
  love.graphics.setFont(defaultFont)
  love.graphics.setColor(255,255,255)
end

function drawPaused()
  love.graphics.setColor(255,0,0)
  love.graphics.setFont(bigFont)
  -- love.graphics.print("GAME OVER",384,300)
  -- love.graphics.print("PAUSE",480,320)
  love.graphics.setFont(defaultFont)
  love.graphics.setColor(255,255,255)
  love.graphics.print("E/S/D/F = movement",10,100)
  love.graphics.print("J = jump",10,115)
  love.graphics.print("K = attack",10,130)
  love.graphics.print("L = grapple",10,145)
  love.graphics.print("I = special",10,160)
end

function drawCalendar()
  ls,lm = getCalendar()
  lsc=16*ls
  lmc=256-(28*lm)
  love.graphics.setColor(lsc,lsc,lsc)
  love.graphics.circle('fill',20,80,10,100)
  love.graphics.setColor(lmc,lmc,lmc)
  love.graphics.circle('fill',40,80,10,100)
  love.graphics.setColor(255,255,255)
  love.graphics.circle('line',20,80,10,100)
  love.graphics.circle('line',40,80,10,100)
  if lm==9 and ls==16 then
    love.graphics.print("eclipse",60,70)
  end
end

function drawDamage()
  love.graphics.setColor(255,255,255)
  love.graphics.translate(-viewport.l,-viewport.t)
  love.graphics.setFont(bigFont)
  for k,v in pairs(damageObjects) do
    if (os.time() - v[4] < 2) then
      love.graphics.print(tostring(v[1]),v[2],v[3])
    end
  end
  love.graphics.setFont(defaultFont)
end

function drawBosspointer()
  --  first check if the boss is offscreen
  if bossx < viewport.l or bossx > viewport.l+viewport.w or bossy < viewport.t or bossy > viewport.t+viewport.h then
    --  calculate
    local px,py = player:getCenter()
    -- bossangle = math.atan2(px-bossx,py-bossy) * 180 / math.pi
    -- bossangle = math.atan2(px-bossx,py-bossy) 
    -- local slope = (bossy - py) / (bossx - px)
    -- local vector = math.sqrt( (bossx-px)^2 + (bossy-py)^2 )
    -- love.graphics.print(bossangle,10,300)
    -- love.graphics.print(slope,10,315)
    -- love.graphics.print(vector,10,330)
    -- love.graphics.print(bossangle,px-10,py-10)
    -- love.graphics.translate(-viewport.l,-viewport.t)
    -- love.graphics.line(bossx,bossy,px,py)
    local vertices = {px,py,bossx,bossy}
    local vectorx = bossx - px
    local vectory = bossy - py
    local dist = math.sqrt(vectorx^2 + vectory^2)
    local point = 80
    vectorx = vectorx/dist
    vectory = vectory/dist
    local middlex = vectorx * 64
    local middley = vectory * 64
    local bigx = vectorx * 72
    local bigy = vectory * 72
    vectorx = vectorx * point
    vectory = vectory * point
    lpx = middlex + 16 * math.cos(30 * math.pi / 180)
    lpy = middley + 16 * math.sin(30 * math.pi / 180)
    local vertices = {px + vectorx, py + vectory, px + lpx, py + lpy, px + middlex, py + middley}
    -- love.graphics.polygon('line',vertices)
    -- love.graphics.line(vertices)
    love.graphics.circle('line',px + vectorx,py + vectory,4,4)
    -- love.graphics.circle('fill',px + bigx,py + bigy,6,100)
    -- love.graphics.circle('fill',px + middlex,py + middley,4,100)
  end

end

function love.draw()                   
   -- first draw untranslated stuff --
  camera.draw(drawCameraStuff)

  love.graphics.setColor(255, 255, 255)
  if gameStart then  --  not started yet, draw the start menu
    menu:draw("start")
  else               --  game started, draw game
    -- local msg = diagnostics:format(tostring(drawDebug), tostring(player.canFly))
    local x,y=player:getCenter()
    local msg = diagnostics:format(collectgarbage("count"),love.timer.getDelta(),love.timer.getFPS(),x,y,player.vx,player.vy)
    love.graphics.print(msg, (love.graphics.getWidth() -150), 10)
    love.graphics.print(("%s | xp: %d | kills: %d | combo: %d | level: %d | stage: %d"):format(player.name,player.score["exp"],player.score["kills"],player.score["combo"],player.score["level"],map.level), 10, 50)
    drawLifebar(player.score["life"],player.score["maxlife"])
    drawPowerbar(player.score["power"],player.score["maxpower"])
    drawCalendar()
    -- ls,lm = getCalendar()
    -- love.graphics.print(("sun: %s | moon: %s"):format(tostring(ls),tostring(lm)), 10, 65)
    if drawDebug then
    printcommands = 100
    if cc then
      for k,v in pairs(cc) do
        if v then love.graphics.print(k,printcommands,100) end
        printcommands = printcommands + 20
      end
    end
    if love.joystick.getJoysticks() then
      local joy = love.joystick.getJoysticks()
      pinput = joy[1]
      if pinput then
        for i=1,15 do
          printcommands = printcommands +10
          if pinput:isDown(i) then
            love.graphics.print(i,printcommands,100)
          end
        end
        axe1 = pinput:getAxis(1)
        axe2 = pinput:getAxis(2)
        love.graphics.print(axe1,10,100)
        love.graphics.print(axe2,20,100)
      end
      -- for k,v in pairs(joy) do
      --   love.graphics.print(tostring(v),10,100)
      -- end
    end
    end

    if (player.score["life"] <= 0) then
      drawGameOver()
    end
    if drawDebug then
      local debugOffset = 100
      -- for k,v in pairs(player.buffer) do  -- debug for command buffer
      -- for k,v in pairs(player.status) do  -- debug for player status
      for k,v in pairs(player.status) do   -- debug for player score
        debugOffset = debugOffset + 15
        love.graphics.print( (tostring(k) .. ": " .. tostring(v)), (love.graphics.getWidth() -150), debugOffset)
      end
    end
    if gamePaused then
      menu:draw("pause")
      -- drawPaused()  -- commented out, handle this from menu instead
    end
    -- draw anything that uses translate here --
    if damageObjects then                 -- uses translate
      drawDamage()
    end
    if bossx and bossy then
      drawBosspointer()
    end
  end -- end gameStart

end

-- ----- --
-- Admin --
-- ----- --

function love.keypressed(k)
  if k=="escape" then love.event.quit() end
  if k=="tab"    then drawDebug = not drawDebug end
  if k=="delete" then
    collectgarbage("collect")
  end
  if k=="return" then
    if not gamePaused and not gameStart then hardreset=true reset(gameType) end
  end
  if k=="rshift" then gamePaused = not gamePaused end
  if k=="lshift" then gamePaused = not gamePaused end
  -- if k=="z" then Object:new(player.l,player.t-200,"lvltoken") end
end
